////////////////////////////////////////////////////////
//  Based on pcidebug.dll & pcidegug.sys
//                        Aug 20 1999 kashiwano masahiro
////////////////////////////////////////////////////////

#include "stddef.h"
#include "CpuInfo.h"

//When you build CpuInfoAMD64.sys, please define _AMD64.
//#define _AMD64


NTSTATUS
DriverEntry(
	IN PDRIVER_OBJECT  DriverObject,
	IN PUNICODE_STRING RegistryPath
	)

/*

Return Value:

	STATUS_SUCCESS if the driver initialized correctly, otherwise an error
    indicating the reason for failure.
*/

{
	PLOCAL_DEVICE_INFO pLocalInfo;	// Device extension:
									// local information for each device.
	NTSTATUS Status;
	PDEVICE_OBJECT DeviceObject;
	int i;

#ifdef ENABLEINT3
	_asm {
		INT 3;
	}
#endif

	DriverObject->MajorFunction[IRP_MJ_CREATE]			= Dispatch;
	DriverObject->MajorFunction[IRP_MJ_CLOSE]			= Dispatch;
	DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL]	= Dispatch;
	DriverObject->DriverUnload							= Unload;
	DriverObject->DriverStartIo = NULL;

	// Create our device.
	Status = CreateDevice(
					MSR_DEVICE_NAME,
					MSR_TYPE,
					DriverObject,
					&DeviceObject
				);

	pLocalInfo = (PLOCAL_DEVICE_INFO)DeviceObject->DeviceExtension;
	pLocalInfo->DeviceObject	= DeviceObject;
	pLocalInfo->DeviceType		= MSR_TYPE;
	for(i=0;i<N_ISR;i++){
		pLocalInfo->table[i].SetupOK = FALSE;
		pLocalInfo->table[i].Handle = NULL;
	}

    return Status;
}

NTSTATUS
CreateDevice(
	IN	PWSTR				PrototypeName,
	IN	DEVICE_TYPE			DeviceType,
	IN	PDRIVER_OBJECT		DriverObject,
	OUT	PDEVICE_OBJECT		*ppDevObj
	)
{
	NTSTATUS Status;						// Status of utility calls
	UNICODE_STRING NtDeviceName;
	UNICODE_STRING Win32DeviceName;

	// Get UNICODE name for device.
	RtlInitUnicodeString(&NtDeviceName, PrototypeName);

	Status = IoCreateDevice(							// Create it.
					DriverObject,
					sizeof(LOCAL_DEVICE_INFO),
					&NtDeviceName,
					DeviceType,
					0,
					TRUE,								// Not Exclusive
					ppDevObj
					);

	if (!NT_SUCCESS(Status))
		return Status;		// Give up if create failed.

	// Clear local device info memory
	RtlZeroMemory((*ppDevObj)->DeviceExtension, sizeof(LOCAL_DEVICE_INFO));

	RtlInitUnicodeString(&Win32DeviceName, DOS_DEVICE_NAME);

	Status = IoCreateSymbolicLink( &Win32DeviceName, &NtDeviceName );

	if (!NT_SUCCESS(Status))	// If we we couldn't create the link then
	{							//  abort installation.
		IoDeleteDevice(*ppDevObj);
	}

	return Status;
}

NTSTATUS
Dispatch(
	IN	PDEVICE_OBJECT pDO,
	IN	PIRP pIrp
	)

/*++

Routine Description:
    This routine is the dispatch handler for the driver.  It is responsible
    for processing the IRPs.

Arguments:
    
    pDO - Pointer to device object.

    pIrp - Pointer to the current IRP.

Return Value:

    STATUS_SUCCESS if the IRP was processed successfully, otherwise an error
    indicating the reason for failure.

--*/

{
	PLOCAL_DEVICE_INFO pLDI;
	PIO_STACK_LOCATION pIrpStack;
	NTSTATUS Status;
	int index;
	//  Initialize the irp info field.
	//      This is used to return the number of bytes transfered.

	pIrp->IoStatus.Information = 0;
	pLDI = (PLOCAL_DEVICE_INFO)pDO->DeviceExtension;	// Get local info struct

    pIrpStack = IoGetCurrentIrpStackLocation(pIrp);

	//  Set default return status
	Status = STATUS_NOT_IMPLEMENTED;

	// Dispatch based on major fcn code.

	switch (pIrpStack->MajorFunction)
	{
		case IRP_MJ_CREATE:
		case IRP_MJ_CLOSE:
			Status = STATUS_SUCCESS;
			break;
		case IRP_MJ_DEVICE_CONTROL:
			//  Dispatch on IOCTL
			switch (pIrpStack->Parameters.DeviceIoControl.IoControlCode)
			{

			case IOCTL_READ_MSR:
				Status = ReadMSR(
						pIrpStack->Parameters.DeviceIoControl.IoControlCode,
						pIrp->AssociatedIrp.SystemBuffer,
						pIrpStack->Parameters.DeviceIoControl.InputBufferLength,
						pIrp->AssociatedIrp.SystemBuffer,
						pIrpStack->Parameters.DeviceIoControl.OutputBufferLength,
						&pIrp->IoStatus.Information);
						//memcpy( pIrp->AssociatedIrp.SystemBuffer ,test,4);
				break;
			case IOCTL_WRITE_MSR:
				Status = WriteMSR(
						pIrpStack->Parameters.DeviceIoControl.IoControlCode,
						pIrp->AssociatedIrp.SystemBuffer,
						pIrpStack->Parameters.DeviceIoControl.InputBufferLength,
						pIrp->AssociatedIrp.SystemBuffer,
						pIrpStack->Parameters.DeviceIoControl.OutputBufferLength,
						&pIrp->IoStatus.Information);
						//memcpy( pIrp->AssociatedIrp.SystemBuffer ,test,4);
				break;
			case IOCTL_HLT:
//				_asm { hlt }
				break;
			}
			break;
	}

	// We're done with I/O request.  Record the status of the I/O action.
	pIrp->IoStatus.Status = Status;

	// Don't boost priority when returning since this took little time.
	IoCompleteRequest(pIrp, IO_NO_INCREMENT );

	return Status;
}

VOID
Unload(
    PDRIVER_OBJECT DriverObject
    )

/*++

Routine Description:
    This routine prepares our driver to be unloaded.  It is responsible
    for freeing all resources allocated by DriverEntry as well as any 
    allocated while the driver was running.  The symbolic link must be
    deleted as well.

Arguments:
    
    DriverObject - Pointer to driver object created by the system.

Return Value:

    None

--*/

{
    PLOCAL_DEVICE_INFO pLDI;
    CM_RESOURCE_LIST NullResourceList;
    BOOLEAN ResourceConflict;
    UNICODE_STRING Win32DeviceName;
	int i;

    // Find our global data
    pLDI = (PLOCAL_DEVICE_INFO)DriverObject->DeviceObject->DeviceExtension;


    RtlInitUnicodeString(&Win32DeviceName, DOS_DEVICE_NAME);

    IoDeleteSymbolicLink(&Win32DeviceName);
	for(i=0;i<N_ISR;i++){
		if(pLDI->table[i].Handle) {
			ZwClose (pLDI->table[i].Handle);
			pLDI->table[i].Handle = NULL;
		}
	}
    IoDeleteDevice(pLDI->DeviceObject);
}

NTSTATUS
ReadMSR(	ULONG IoctlCode,
			void *lpInBuffer, 
			ULONG nInBufferSize, 
			void *lpOutBuffer, 
			ULONG nOutBufferSize, 
			ULONG_PTR* lpBytesReturned)
{
	ULONG ulECX = *(ULONG*)lpInBuffer;
	ULONG ulEAX = 0;
	ULONG ulEDX = 0;

#ifndef _AMD64
	_asm{
		mov ecx,ulECX
		rdmsr
		mov ulEAX,eax
		mov ulEDX,edx
	}
	memcpy((PULONG)lpOutBuffer,&ulEAX,4);
	memcpy((PULONG)lpOutBuffer+1,&ulEDX,4);
#else
	ULONG64 data = __readmsr(ulECX);
	memcpy((PULONG)lpOutBuffer,&data,8);
#endif

	*lpBytesReturned = 8;
	return STATUS_SUCCESS;
}

NTSTATUS
WriteMSR(	ULONG IoctlCode,
			void *lpInBuffer, 
			ULONG nInBufferSize, 
			void *lpOutBuffer, 
			ULONG nOutBufferSize, 
			ULONG_PTR* lpBytesReturned)
{

	ULONG ulECX = *(ULONG*)(PULONG)lpInBuffer;
#ifndef _AMD64
	ULONG ulEAX = *(ULONG*)((PULONG)lpInBuffer + 1);
	ULONG ulEDX = *(ULONG*)((PULONG)lpInBuffer + 2);

	_asm{
		mov eax,ulEAX
		mov edx,ulEDX
		mov ecx,ulECX
		wrmsr
	}
	*lpBytesReturned = 0;
#else
	ULONG64 ulEAX = *(ULONG*)((PULONG)lpInBuffer + 1);
	ULONG64 ulEDX = *(ULONG*)((PULONG)lpInBuffer + 2);
	ULONG64 data = ( ulEDX << 32 ) + ulEAX;
	__writemsr(ulECX, data);
#endif

	return STATUS_SUCCESS;
}
