/*---------------------------------------------------------------------------*/
//       Author : hiyohiyo
//         Mail : hiyohiyo@crystalmark.info
//          Web : http://crystalmark.info/
//      License : The modified BSD license
//
//                                Copyright 2002- hiyohiyo All rights reserved.
/*---------------------------------------------------------------------------*/

#include "stdafx.h"
#include "CpuUsage.h"

static int GetCpuUsageNT(int number);
static int GetCpuUsage9x();
static BOOL FlagNT = FALSE;
static BOOL IsNT();

typedef LONG (WINAPI *__NTQSI__)(UINT,PVOID,ULONG,PULONG);

static __NTQSI__ NTQSI = NULL;
static HMODULE hModule = NULL;
static DWORD NumberOfProcessors = 1;

static HKEY hKey, hDummy;
static DWORD dummy, dummysize=sizeof(DWORD);
static DWORD CpuUsage, size=sizeof(DWORD);

static void InitCpuUsageNT();
static int GetCpuUsageNT(int number);

static void InitCpuUsage9x();
static int GetCpuUsage9x();
static void CleanupCpuUsage9x();

void InitCpuUsage()
{
	FlagNT = IsNT();
	if( FlagNT ){
		InitCpuUsageNT();
	}else{
		InitCpuUsage9x();
	}
}

int GetCpuUsage(int number)
{
	if( FlagNT ){
		return GetCpuUsageNT(number);
	}else{
		return GetCpuUsage9x();
	}
}

void CleanupCpuUsage()
{
	if( FlagNT ){
		
	}else{
		CleanupCpuUsage9x();
	}
}

static void InitCpuUsageNT()
{
	SYSTEM_INFO si;
	GetSystemInfo(&si);
	NumberOfProcessors = si.dwNumberOfProcessors;
	hModule = GetModuleHandle("ntdll");
	NTQSI = (__NTQSI__)GetProcAddress(
				hModule,
				"NtQuerySystemInformation"
				);
}

static int GetCpuUsageNT(int number)
{
	if( NTQSI == NULL ){
		return -1;
	}
	SYSTEM_PERFORMANCE_INFORMATION SysPerfInfo;
	SYSTEM_TIME_INFORMATION        SysTimeInfo;
    double                         IdleTime;
    double                         SystemTime;
    LONG                           status;
    static LARGE_INTEGER           OldIdleTime   = {0, 0};
    static LARGE_INTEGER           OldSystemTime = {0, 0};

	int CpuUsage = 0;

	status = NTQSI(SystemTimeInformation,&SysTimeInfo,sizeof(SysTimeInfo),0);
	if (status!=NO_ERROR)
		return -1;
	
	status = NTQSI(SystemPerformanceInformation,&SysPerfInfo,sizeof(SysPerfInfo),NULL);
	if (status != NO_ERROR)
		return -1;
	
	if (OldIdleTime.QuadPart != 0)
	{
		IdleTime = (double)(SysPerfInfo.IdleTime.QuadPart - OldIdleTime.QuadPart);
		SystemTime = (double)(SysTimeInfo.SystemTime.QuadPart - OldSystemTime.QuadPart);
		IdleTime = IdleTime / SystemTime;
		IdleTime = 100.0 - IdleTime * 100.0 / NumberOfProcessors + 0.5;
		
		CpuUsage = (int)IdleTime;
	}
	
	OldIdleTime = SysPerfInfo.IdleTime;
	OldSystemTime = SysTimeInfo.SystemTime;

	return (int)CpuUsage;
}

static void InitCpuUsage9x()
{
	RegOpenKey (HKEY_DYN_DATA, "PerfStats\\StatData", &hKey);

	RegOpenKey (HKEY_DYN_DATA, "PerfStats\\StartStat", &hDummy); 
	RegQueryValueEx (hDummy, "KERNEL\\CPUUsage", NULL, NULL, (LPBYTE)&dummy, &dummysize); 
	RegCloseKey (hDummy);
}

static int GetCpuUsage9x()
{
	RegQueryValueEx(hKey,"KERNEL\\CPUUsage",NULL,NULL,(LPBYTE)&CpuUsage,&size);
	return (int)CpuUsage;
}

static void CleanupCpuUsage9x()
{
	RegOpenKeyA(HKEY_DYN_DATA, "PerfStats\\StopStat", &hDummy); 
	RegQueryValueEx(hDummy, "KERNEL\\CPUUsage", NULL, NULL, (LPBYTE)&dummy, &dummysize); 
	RegCloseKey(hDummy);
	
	RegCloseKey(hKey);
}

static BOOL IsNT()
{
	OSVERSIONINFO OSVersionInfo;
	OSVersionInfo.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	GetVersionEx(&OSVersionInfo);

	return OSVersionInfo.dwPlatformId == VER_PLATFORM_WIN32_NT;
}
